# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel


class Folder(UniversalBaseModel):
    """
    Response model for a folder
    """

    title: str = pydantic.Field()
    """
    Folder title
    """

    description: typing.Optional[str] = pydantic.Field(default=None)
    """
    Folder description
    """

    folder_id: str
    organization_id: str
    workflow_count: typing.Optional[int] = pydantic.Field(default=None)
    """
    Number of workflows in this folder
    """

    created_at: dt.datetime
    modified_at: dt.datetime

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow
