Original prompt:

```
{{ prompt }}
```
{% if context %}

Additional context about the user's needs:
```
{{ context }}
```
{% endif %}

Can you improve the original prompt for an AI browser agent?

Respond ONLY with valid JSON in this format with no additional text before or after it:
```json
{
  "improved_prompt": str, // The improved version of the prompt
}

Ensure that the "improved_prompt" contains liberal whitespace tokens for formatting, clarity, and legibility.
